<?php

class EveryBoard_Renderer {

    private $main_site_id;
    private $active_device;
    private $eb_fw_helper;
	private $env_settings;
	private static $widget_ids      = array();
	private static $widget_counter  = 0;

	private $max_cols;
	private $nested_col_sizes       = array();
	private $nested_board_counter   = 0;


    /**
     * Class Constructor
     */
    public function __construct() {
        $this->active_device    = $this->set_active_device();
        $this->eb_fw_helper     = new EveryBoard_Framework_Helper( false );
	    $this->env_settings     = new EnvSettings();
	    $this->max_cols         = $this->eb_fw_helper->eb_max_columns;
        $this->main_site_id     = get_current_blog_id();
    }

    /*
	 * Function to Initialize the rendering of a Board
	 * Used in "The Content Filter" to replace the WP-content with Board Content
	 *
	 * @param $board_json
	 * @return Board HTML
	 */
    public function render( $board_json, $site_id = null ) {

        $board_json_arr = json_decode( $board_json[0], true );
        $board_json     = json_decode( $board_json[0] );

        // Fetch all articles for board in a single request.
        EveryBoard_Article_Helper::prefetch_board_articles( $board_json_arr );

        $return_data    = '<section class="'.$this->eb_fw_helper->board_wrapper_classes().'">';

        if ( isset( $board_json->rows ) ) {
            foreach ( $board_json->rows as $row ) {
                $return_data .= $this->render_structure_item( $row, $site_id );
            }
        }

        $return_data    .= '</section>';
        return $return_data;
    }

    /**
     * Renders a structure item.
     *
     * @param $item
     * @param $site_id
     * @return string
     */
    private function render_structure_item( $item, $site_id = null, $page_col_size = null ) {

        if( $this->should_item_render( $item ) ) {

            $return_html = "";
            $css = $this->generate_item_css( $item->settings );

            if( $page_col_size !== null && $item->type !== 'row' && $item->type !== 'col' && $item->type !== 'linked_widget' ) {
              $css .= ' ' . $this->eb_fw_helper->get_content_size_classes( $page_col_size );
            }

            if( isset( $item->type ) ) {
                switch( $item->type ) {
                    case 'row':
                        $return_html .= $this->render_row( $item, $css, $site_id );
                        break;
                    case 'col':
                        $return_html .= $this->render_col( $item, $css, $site_id );
                        break;
                    case 'widget':
                        $return_html .= $this->render_widget( $item, $css, $site_id );
                        break;
                    case 'oc_article';
                        $return_html .= $this->render_article( $item, $css );
                        break;
                    case 'linked_widget';
                        $return_html .= $this->render_linked_widget( $item, $css, $site_id );
                        break;
                    case "embed_widget":
                        $return_html .= $this->render_embed_widget( $item, $css );
                        break;
                    case "articlelist_widget":
                        $return_html .= $this->render_articlelist_widget( $item, $css, $site_id );
                        break;
                }
            }

            return $return_html;
        }
    }

    /**
     * Renders a row.
     *
     * @param $item
     * @param $css
     * @return string
     */
    private function render_row( $item, $css, $site_id = null ) {

        $return_html    = "";
        $return_html    .= $this->eb_fw_helper->row_html( $css );

        if ( isset( $item->cols ) ) {
            foreach ( $item->cols as $col ) {
                $return_html .= $this->render_structure_item( $col, $site_id );
            }
        }

        $return_html .= '<div class="clearfix"></div></div>';

        return $return_html;
    }

	private function calculate_page_col_size($nested_col_sizes) {

		# Remove the top layer and use as start values.
		$page_col_size = array_shift($nested_col_sizes);

		# Iterate threw the nested boards column sizes
		foreach ( $nested_col_sizes as $board_col_size ) {

			# Iterate threw the device-specific sizes
			foreach ( $board_col_size as $device => $col_size) {

				$col_size = (int)$col_size; # make sure we are using int:s

				# Only recalculate if the column is less than max-size
				if ($col_size < $this->max_cols) {

					$col_size_in_percent = $col_size / $this->max_cols;
					$new_size = round($page_col_size[$device] * $col_size_in_percent, 0);
					$page_col_size[$device] = ($new_size < 1) ? 1 : $new_size;
				}
			}
		}
		return $page_col_size;
	}

    /**
     * Renders a column.
     *
     * @param $item
     * @param $css
     * @return string
     */
    private function render_col( $item, $css, $site_id = null ) {

	    # Save the sizes of the current column
	    $current_col_size = array(
		    'mobile' => $item->settings->devices->mobile->colspan,
		    'tablet' => $item->settings->devices->tablet->colspan,
		    'smallscreen' => $item->settings->devices->smallscreen->colspan,
		    'largescreen' => $item->settings->devices->largescreen->colspan
	    );
	    $this->nested_col_sizes[$this->nested_board_counter] = $current_col_size;
	    $page_col_size = ($this->nested_board_counter > 0) ? $this->calculate_page_col_size($this->nested_col_sizes) : $current_col_size;

	    $return_html = $this->eb_fw_helper->col_html( $current_col_size, $css, $page_col_size );

	    if( !isset( $item->settings->hiddenwrapper ) || $item->settings->hiddenwrapper !== "true" ) {
		    $return_html .= '<div class="' . $css . '">';
	    }

        if ( isset( $item->content ) ) {
            foreach ( $item->content as $content ) {

                if(isset($content->settings) && count($content->settings) > 0 && is_string($content->settings)) {
                    #$content->settings = EveryBoard_Settings::decode_settings_array( $content->settings );
                    $content->settings = json_decode($content->settings); // TODO: WTF?!
                }

                // Inherit the settings from content if none present
                if( ( !isset( $content->settings->layout ) && isset( $item->settings->layout ) ) || ( isset( $content->settings->layout ) && $content->settings->layout == "None" && isset( $item->settings->layout ) ) ) {

                    if( is_object( $content->settings ) ) {
                        $content->settings->layout = $item->settings->layout;
                    }
                }

                $return_html .= $this->render_structure_item( $content, $site_id, $page_col_size );
            }
        }

        if( !isset( $item->settings->hiddenwrapper ) || $item->settings->hiddenwrapper !== "true" ) {
            $return_html .= '</div>';
            $return_html .= '</div>';
        } else {
            $return_html .= '</div>';
        }

        return $return_html;
    }

    /**
     * Renders an article.
     *
     * @param $item
     * @param $css
     * @return string
     */
    private function render_article( $item, $css ) {

        $article        = EveryBoard_Article_Helper::get_oc_article( $item->oc_uuid );
		$article_html   = '';
	    $should_render  = true;

	    if( is_a( $article, 'OcArticle' ) && $article !== null ) {
		    // If articles with pubdates in the future should be hidden.
		    if ( $this->env_settings->get_pubdate_hide_property() && isset( $article->pubdate[0] ) ) {
			    $pubdate_timestamp = strtotime( $article->pubdate[0] );
			    $now_timestamp     = time();

			    if ( $pubdate_timestamp > $now_timestamp ) {
				    $should_render = false;
			    }
		    }

		    if( $this->env_settings->get_use_trashed() && isset( $article->everytrashed ) && is_array( $article->everytrashed ) ) {

			    if( in_array( OcUtilities::get_site_shortname(), $article->everytrashed ) ) {
					$should_render = false;
			    }
		    }

		    if ( $should_render ) {

			    $article_html .= '<div class="rendered_board_article ' . $css . '">';
			    if ( isset( $item->settings->layout ) && $item->settings->layout !== "None" && file_exists( EveryBoard_Settings::get_template_path( $item->settings->layout ) ) ) {
				    ob_start();
				    include EveryBoard_Settings::get_template_path( $item->settings->layout );
				    $article_html .= ob_get_contents();
				    ob_end_clean();
			    } else if ( file_exists( EveryBoard_Settings::get_template_path( "templates/ocarticle-default.php" ) ) ) {
				    ob_start();
				    include EveryBoard_Settings::get_template_path( "templates/ocarticle-default.php" );
				    $article_html .= ob_get_contents();
				    ob_end_clean();
			    } else {
				    $article_html .= $this->render_article_template_fallback( $article );
			    }

			    $article_html .= '</div>';
		    }

		    return $article_html;
	    }
	}

    private function render_article_object( $article, $item, $css ) {
        $article_html   = '';
        $should_render  = true;

        if( is_a( $article, 'OcArticle' ) && $article !== null ) {
            // If articles with pubdates in the future should be hidden.
            if ( $this->env_settings->get_pubdate_hide_property() && isset( $article->pubdate[0] ) ) {
                $pubdate_timestamp = strtotime( $article->pubdate[0] );
                $now_timestamp     = time();

                if ( $pubdate_timestamp > $now_timestamp ) {
                    $should_render = false;
                }
            }

            if( $this->env_settings->get_use_trashed() && isset( $article->everytrashed ) && is_array( $article->everytrashed ) ) {

                if( in_array( OcUtilities::get_site_shortname(), $article->everytrashed ) ) {
                    $should_render = false;
                }
            }

            if ( $should_render ) {

                $article_html .= '<div class="rendered_board_article' . $css . '">';
                if ( isset( $item->settings->layout ) && $item->settings->layout !== "None" && file_exists( EveryBoard_Settings::get_template_path( $item->settings->layout ) ) ) {
                    ob_start();
                    include EveryBoard_Settings::get_template_path( $item->settings->layout );
                    $article_html .= ob_get_contents();
                    ob_end_clean();
                } else if ( file_exists( EveryBoard_Settings::get_template_path( "templates/ocarticle-default.php" ) ) ) {
                    ob_start();
                    include EveryBoard_Settings::get_template_path( "templates/ocarticle-default.php" );
                    $article_html .= ob_get_contents();
                    ob_end_clean();
                } else {
                    $article_html .= $this->render_article_template_fallback( $article );
                }

                $article_html .= '</div>';
            }

            return $article_html;
        }
    }

    /**
     * Render a widget.
     *
     * @param $widget
     * @param $css
     * @param $site_id
     * @return string
     */
    public static function render_widget( $widget, $css, $site_id = null ) {

        $content = "";
        if( isset( $widget->widget_id ) ) {

            $switched           = false;
            $original_blog_id   = get_current_blog_id();
            if( isset( $site_id ) ) {

                if( $original_blog_id !== $site_id ) {
                    switch_to_blog( $site_id );
                    $switched = true;
                }
            }

            $widget_instance = EveryBoard_WidgetPrefab::get_widget_instance( $widget->widget_id, $switched );

	        // Check if the widget has already been rendered, in that case prepend it's id to make it unique.
	        if( in_array( $widget->widget_id, self::$widget_ids ) ) {
		        $widget->widget_id = 'eb' . self::$widget_counter . '_' . $widget->widget_id;
		        self::$widget_counter++;
	        }
	        else {
		        array_push( self::$widget_ids, $widget->widget_id );
	        }

            if( $switched ) {
                restore_current_blog();
            }


            if($widget_instance === null) {
                return "";
            }
            $widget_instance['settings']['id'] = $widget->widget_id;

            ob_start();
            echo '<div class="rendered_board_widget'.$css.'" id="' . $widget->widget_id . '">';
            the_widget( $widget_instance['classname'], $widget_instance['settings'] );
            echo '</div>';
            $content = ob_get_contents();

            ob_end_clean();
        }

        return $content;
    }

    /**
     * Renders an embed widget.
     *
     * @param $widget
     * @param $css
     * @return string
     */
    private function render_embed_widget( $widget, $css ) {

        $content = "";

        if( isset( $widget->settings->content ) ) {

            ob_start();
            echo '<div class="rendered_board_widget'.$css.'">';
            echo do_shortcode( base64_decode( $widget->settings->content ) );
            echo '</div>';
            $content = ob_get_contents();

            ob_end_clean();
        }

        return $content;
    }

    /**
     * Renders a linked board inside another board.
     *
     * @param $board
     * @param $css
     * @return string
     */
    private function render_linked_widget( $board, $css, $site_id = null ) {

        $content = "";
        if( isset( $board->settings->board_id ) && $board->settings->board_id !== "0" ) {
	        $this->nested_board_counter++;

	        $original_blog_id = get_current_blog_id();
	        if( isset( $site_id ) ) {
		        switch_to_blog( $site_id );
	        }

            ob_start();
            echo '<div class="rendered_board_linked_widget'.$css.'">';

            // Use the temp json if in preview
            if ( isset( $_GET["board_preview"] ) && $_GET["board_preview"] == "true" ) {
                $board_json = get_post_custom_values( 'everyboard_json_temp', $board->settings->board_id );
            } else {
                $board_json = get_post_custom_values( 'everyboard_json', $board->settings->board_id );
            }

            echo $this->render( $board_json );

            echo '</div>';
            $content = ob_get_contents();

            ob_end_clean();

	        if( isset( $site_id ) ) {
                restore_current_blog();
	        }

	        unset($this->nested_col_sizes[$this->nested_board_counter]);
	        $this->nested_board_counter--;
        }
        else if( isset( $board->settings->network_board_id, $board->settings->network_board_site_id )
                 && $board->settings->network_board_id !== "0" && $board->settings->network_board_site_id !== "0" ) {

            $this->nested_board_counter++;
            $board_id           = intval( $board->settings->network_board_id );
            $site_id            = intval( $board->settings->network_board_site_id );
            $original_blog_id   = get_current_blog_id();

            $render_on_source   = false;
            if( isset( $board->settings->network_board_source_render ) && $board->settings->network_board_source_render === 'true' ) {
                $render_on_source = true;
            }
            //global $wp_widget_factory;

            switch_to_blog( $site_id );

            $board_json = get_post_custom_values( 'everyboard_json', $board_id );

            // If the board should NOT be rendered on source site change back to the current site BEFORE render.
            if( !$render_on_source ) {
                restore_current_blog();
            }

            ob_start();
            echo '<div class="rendered_board_linked_widget'.$css.'">';
            echo $this->render( $board_json, $site_id );
            echo '</div>';
            $content = ob_get_contents();
            ob_end_clean();

            // If the board should be rendered on source site change back to the current site AFTER render.
            if( $render_on_source ) {
                restore_current_blog();
            }

            // We have to reset the nested board counter after we used it.
            unset($this->nested_col_sizes[$this->nested_board_counter]);
            $this->nested_board_counter--;
        }

        return $content;
    }

    private function render_articlelist_widget( $item, $css, $site_id = null ) {
        $content = "";

        if( isset($item->settings->list) && $item->settings->list !== '0' ) {
            $list = (int)$item->settings->list;
            $listposition = (int)$item->settings->listposition;
            $limit = isset( $item->settings->listlimit ) ? (int)$item->settings->listlimit : 1;
            $listarticle = null;

            $switched = false;
            $original_blog_id   = get_current_blog_id();

            if( isset( $site_id ) ) {
                if( $original_blog_id !== $site_id ) {
                    $switched = true;
                    switch_to_blog( $site_id );
                }
            }

            if( $limit > 1 ) {
              $listarticles = EveryList::get_articles_from_list($list, $listposition, $limit);
            }
            else {
              $listarticle = EveryList::get_article_from_list($list, $listposition);
            }


            if( $switched ) {
                restore_current_blog();
            }

            if( $listarticle !== null ) {
                if( $switched || get_current_blog_id() !== $this->main_site_id ) {
                    $article = EveryBoard_Article_Helper::force_fetch_oc_article( $listarticle->uuid[0] );
                    $content = $this->render_article_object($article, $item, $css);
                } else {
                    $item->oc_uuid = $listarticle->uuid[0];
                    $content = $this->render_article($item, $css);
                }
            }
            else {
              $content = "";
              if( $switched || get_current_blog_id() !== $this->main_site_id ) {

                  foreach( $listarticles as $larticle ) {
                    $article = EveryBoard_Article_Helper::force_fetch_oc_article( $larticle->get_value('uuid') );
                    $content .= $this->render_article_object($article, $item, $css);
                  }
              } else {
                  foreach( $listarticles as $article ) {
                    $item->oc_uuid = $article->get_value('uuid');
                    $content .= $this->render_article($item, $css);
                  }
              }
            }


        } else if( isset($item->settings->network_list, $item->settings->network_list_site_id) &&
            $item->settings->network_list !== 0 && $item->settings->network_list_site_id !== '' ) {

            $list = (int)$item->settings->network_list;
            $listposition = (int)$item->settings->listposition;
            $limit = isset( $item->settings->listlimit ) ? (int)$item->settings->listlimit : 1;
            $network_list_site = (int)$item->settings->network_list_site_id;
            $listarticle = null;

            $switched = false;
            $inner_switched = false;
            $current_blog_id = get_current_blog_id();

            if( isset($site_id) && $current_blog_id !== $site_id ) {
                switch_to_blog($site_id);
                $switched = true;
            }

            if( $network_list_site !== get_current_blog_id() ) {
                switch_to_blog( $network_list_site );
                $inner_switched = true;
            }

            if( $limit > 1 ) {
              $listarticles = EveryList::get_articles_from_list($list, $listposition, $limit);
            }
            else {
              $listarticle = EveryList::get_article_from_list($list, $listposition);
            }

            if( $inner_switched ) {
                restore_current_blog();
            }

            if( $switched ) {
                restore_current_blog();
            }

            if( $listarticle !== null ) {
                if( $switched || get_current_blog_id() !== $this->main_site_id ) {
                    $article = EveryBoard_Article_Helper::force_fetch_oc_article( $listarticle->uuid[0] );
                    $content = $this->render_article_object($article, $item, $css);
                } else {
                    $item->oc_uuid = $listarticle->uuid[0];
                    $content = $this->render_article($item, $css);
                }
            }
            else {
              $content = "";
              if( $switched || get_current_blog_id() !== $this->main_site_id ) {

                  foreach( $listarticles as $larticle ) {
                    $article = EveryBoard_Article_Helper::force_fetch_oc_article( $larticle['uuid'] );
                    $content .= $this->render_article_object($article, $item, $css);
                  }
              } else {
                  foreach( $listarticles as $article ) {
                    $item->oc_uuid = $article['uuid'];
                    $content .= $this->render_article($item, $css);
                  }
              }
          }
        }

        return $content;
    }

    /**
     * Fallback for article render if no template is found use this general article render.
     *
     * @param $article
     * @return string
     */
    public function render_article_template_fallback( $article ) {

        $article_img    = isset( $article->imageuuid[0] ) ? $article->imageuuid[0] : null;
        $article_title  = isset( $article->headline[0] ) ? $article->headline[0] : "";
        $artile_leadin  = isset( $article->leadin[0] ) ? $article->leadin[0] : null;
        $article_text   = isset( $article->text[0] ) ? $article->text[0] : "";
        $article_author = isset( $article->author[0] ) ? $article->author[0] : "";

        $article_html   = "";

        if ( isset( $article_img ) ) {
            $image_src = plugins_url( 'Every/image.php' ) . '?uuid=' . $article_img . '&type=preview&source=false&function=cover';
            $article_html .= '<img src="' . $image_src . '" style="width: 100%; height: auto;" />';
        }

        if ( isset( $article_title ) ) {
            $article_html .= '<h2>' . $article_title . '</h2>';
        }

        if ( isset( $artile_leadin ) ) {
            $article_html .= '<h4>' . $artile_leadin . '</h4>';
        }

        if ( isset( $article_author ) ) {
            $article_html .= '<p>' . $article_author . '</p>';
        }

        if ( isset( $article_text ) ) {
            $article_html .= '<div>' . OcUtilities::truncate_article_text( $article_text, 200 ) . '</div>';
        }

        return $article_html;
    }

  	/**
	 * Uses Mobile Detect library to check what device type the current user is browsing from.
	 *
	 * @return string
	 */
	private function set_active_device() {

		if( isset( $_SERVER['EVERY_DEVICE'] ) ) {
			return $_SERVER['EVERY_DEVICE'];
		}
		else {
			return "mobile";
		}
	}

    /**
     * Checks the items settings to see if the item should be rendered on current device.
     *
     * @param $item
     * @return bool
     */
    private function should_item_render( $item ) {

        $should_render = true;
        if( isset( $item->settings->devices ) ) {

            if( $this->active_device === 'mobile' && $item->settings->devices->mobile->removed === 'true' ) {
                $should_render = false;
            }
            else if( $this->active_device === 'tablet' && $item->settings->devices->tablet->removed === 'true' ) {
                $should_render = false;
            }
            else if( $this->active_device === 'desktop' && $item->settings->devices->smallscreen->removed === 'true' ||
                    $this->active_device === 'desktop' && $item->settings->devices->largescreen->removed === 'true' ) {
                $should_render = false;
            }
        }

        return $should_render;
    }

    /**
     * Generates custom CSS for an items settings.
     *
     * @param $settings
     * @return string
     */
    private function generate_item_css( $settings ){

        $custom_css = ' ';

        // Loop over any existing css classes from settings.
        if( isset( $settings->cssclass ) && is_array( $settings->cssclass ) ) {
            foreach ($settings->cssclass as $key => $value) {
                $custom_css .= $value . ' ';
            }
        }
        else if( isset( $settings->cssclass ) && is_string( $settings->cssclass ) ) {
            $custom_css .= $settings->cssclass . ' ';
        }

        // Set any framework specific classes for the settings.
        $custom_css .= $this->eb_fw_helper->get_item_css_classes( $settings );
        return $custom_css;
    }
}