<?php

  class Everyboard_Dashboard {

    private $versions_url = 'https://composer.infomaker.io/versions.json';

    public function __construct() {

      add_action( 'wp_dashboard_setup', [ &$this, 'eb_add_dashboard_widget' ] );
    }

    /**
     * Add a widget to the admin dashboard.
     */
    function eb_add_dashboard_widget() {
      wp_add_dashboard_widget(
        'eb_dashboard_widget',
        'Everyboard',
        [ &$this, 'eb_dashboard_widget_render' ]
      );
    }

    /**
     * Outputs data to dashboard widget.
     */
    function eb_dashboard_widget_render() {

      $key = 'eb_latest_version';
      if( false === ( $latest_version = get_transient( $key ) ) ) {
        $response = wp_remote_get( $this->versions_url, [
          'timeout' => 3
        ] );

        if( is_array($response) ) {
          $plugins = json_decode( $response['body'] );
          if( $plugins !== null ) {
            foreach( $plugins as $plugin ) {
              if( $plugin->name === 'infomaker/everyboard' ) {
                usort($plugin->versions, 'version_compare');
                $latest_version = $plugin->versions[count($plugin->versions)-1];
                set_transient( $key, $latest_version, 86400 );
              }
            }
          }
        }
      }

      print '<p>';
        print __('Currently running Everyboard version: ', 'everyboard') . '<strong>' . EVERYBOARD_VERSION . '</strong>';
      print '</p>';

      if( $latest_version === false ) {
        print '<p>';
          print __('Could not fetch version information, check back later!', 'everyboard');
        print '</p>';
      }
      else if( $latest_version === EVERYBOARD_VERSION ) {
        print '<p>';
          print __('Up to date with latest release.', 'everyboard');
        print '</p>';
      }
      else if( version_compare( EVERYBOARD_VERSION, $latest_version, '<' ) ) {
        print '<p>';
          print __('There is a newer version of Everyboard available, update or contact Infomaker for more information.', 'everyboard') . ' ';
          print '<strong>' . __('Latest version', 'everyboard') . ': ' . $latest_version . '</strong>';
        print '</p>';
      }

    }
  }